require 'minitest/autorun'
require 'pepito/robot'

class TestPepitoRobot < MiniTest::Test
  def setup
    ENV['LOG_FILE'] = '/dev/null'

    @mock_http_api = MiniTest::Mock.new
    @mock_database = MiniTest::Mock.new
    @mock_database.expect(:get, nil, ['name'])

    Pepito::HTTPApi::RackApp.stub :new, @mock_http_api do
      Pepito::Database.stub :new, @mock_database do
        @robot = Pepito::Robot.new
      end
    end
  end

  def test_attributes
    assert_equal 'pepito', @robot.name
    assert_equal Pepito::WebApp::App, @robot.web_app
    assert_instance_of Logger, @robot.logger
    assert_instance_of Hash, @robot.handlers
    assert_instance_of Hash, @robot.adapters
    assert_instance_of Array, @robot.default_adapters
    assert_instance_of Array, @robot.default_handlers
    assert_empty @robot.handlers
    assert_empty @robot.adapters
  end

  def test_methods
    assert_respond_to @robot, :run
    assert_respond_to @robot, :receive_message
    assert_respond_to @robot, :start_handler
    assert_respond_to @robot, :stop_handler
    assert_respond_to @robot, :handler?
    assert_respond_to @robot, :handlers_list
    assert_respond_to @robot, :start_adapter
    assert_respond_to @robot, :stop_adapter
    assert_respond_to @robot, :adapter?
    assert_respond_to @robot, :adapters_list
  end

  def test_handlers_list
    assert_instance_of Array, @robot.handlers_list
  end

  def test_adapters_list
    assert_instance_of Array, @robot.adapters_list
  end

  def test_start_handler
    # Fail to start if handler is running
    @robot.handlers.stub :key?, true do
      assert_equal 1, @robot.start_handler('', 'pepito', {})
    end

    # Fail to start if handler does not exist
    @robot.handlers.stub :key?, false do
      assert_equal 2, @robot.start_handler('', 'pepito', {})
    end

    # Fail with other errors
    @robot.handlers.stub :key?, false do
      mock_handler = Class.new do
        def initialize(_robot, _config)
          raise 'error'
        end
      end
      Pepito::Handlers.stub :const_get, mock_handler do
        assert_equal 4, @robot.start_handler('', 'pepito', {})
      end
    end

    # Start handler successfully
    @robot.handlers.stub :key?, false do
      name = 'handler_name'
      gem_name = 'pepito'
      config = { 'name' => 'option' }

      mock_handler = MiniTest::Mock.new
      mock_handler.expect(:run, nil)
      mock_handler.expect(:start, nil)

      mock_handler_klass = MiniTest::Mock.new
      mock_handler_klass.expect(:new, mock_handler, [@robot, config])

      mock_router = MiniTest::Mock.new
      mock_router.expect(:uncompile, nil)

      @mock_http_api.expect(:router, mock_router)
      @mock_database.expect(:sadd, nil, ['handlers', name])
      @mock_database.expect(:set, nil, [name + ':gem', gem_name])
      config.each do |k, v|
        @mock_database.expect(:hset, nil, [name + ':config', k, v])
      end

      Pepito::Handlers.stub :const_get, mock_handler_klass do
        assert_equal 0, @robot.start_handler(name, gem_name, config)
        assert mock_handler.verify
        assert mock_router.verify
        assert mock_handler_klass.verify
        assert @mock_http_api.verify
        assert @mock_database.verify
      end
    end
  end

  def test_stop_handler
    # Fail to stop if handler is in the default handlers list
    @robot.default_handlers.stub :include?, true do
      assert_equal 1, @robot.stop_handler('')
    end

    # Fail to stop if handler is not running
    @robot.default_handlers.stub :include?, false do
      @robot.handlers.stub :key?, false do
        assert_equal 2, @robot.stop_handler('')
      end
    end

    # Stop successfully otherwise
    @robot.default_handlers.stub :include?, false do
      @robot.handlers.stub :key?, true do
        @robot.handlers.stub :delete, nil do
          name = 'handler'
          mock_router = MiniTest::Mock.new
          mock_router.expect(:reset!, nil)
          @mock_database.expect(:srem, nil, ['handlers', name])
          @mock_database.expect(:del, nil, [name + ':config'])
          @mock_database.expect(:del, nil, [name + ':gem'])
          @mock_http_api.expect(:router, mock_router)

          assert_equal 0, @robot.stop_handler(name)
          assert mock_router.verify
          assert @mock_http_api.verify
          assert @mock_database.verify
        end
      end
    end
  end

  def test_handler?
    @robot.handlers.stub :key?, true do
      assert @robot.handler?('')
    end
    @robot.handlers.stub :key?, false do
      refute @robot.handler?('')
    end
  end

  def test_start_adapter
    # Fail to start if handler is running
    @robot.adapters.stub :key?, true do
      assert_equal 1, @robot.start_adapter('', '', {})
    end

    # Fail to start if gem can't be loaded does not exist
    @robot.adapters.stub :key?, false do
      assert_equal 2, @robot.start_adapter('', '', {})
    end

    # Fail to start if gem is loaded but adapter doesn't exist
    @robot.adapters.stub :key?, false do
      assert_equal 2, @robot.start_adapter('', 'pepito', {})
    end

    # Fail if required methods missing
    @robot.adapters.stub :key?, false do
      mock_adapter = Class.new do
        def initialize(_robot, _config)
          raise Pepito::Errors::RequiredMethodsError.new([]), ''
        end
      end
      Pepito::Adapters.stub :const_get, mock_adapter do
        assert_equal 3, @robot.start_adapter('', 'pepito', {})
      end
    end

    # Fail if missing configuration
    @robot.adapters.stub :key?, false do
      mock_adapter = Class.new do
        def initialize(_robot, _config)
          raise Pepito::Errors::MissingConfigurationValueError.new([]), ''
        end
      end
      Pepito::Adapters.stub :const_get, mock_adapter do
        assert_equal 4, @robot.start_adapter('', 'pepito', {})
      end
    end

    # Fail with other errors
    @robot.adapters.stub :key?, false do
      mock_adapter = Class.new do
        def initialize(_robot, _config)
          raise 'error'
        end
      end
      Pepito::Adapters.stub :const_get, mock_adapter do
        assert_equal 5, @robot.start_adapter('', 'pepito', {})
      end
    end

    # Start adapter successfully
    @robot.adapters.stub :key?, false do
      name = 'adapter_name'
      gem_name = 'pepito'
      config = { 'name' => 'option' }

      mock_adapter = MiniTest::Mock.new
      mock_adapter.expect(:run, nil)

      mock_adapter_klass = MiniTest::Mock.new
      mock_adapter_klass.expect(:new, mock_adapter, [@robot, config])

      @mock_database.expect(:sadd, nil, ['adapters', name])
      @mock_database.expect(:del, nil, [name])
      @mock_database.expect(:set, nil, [name + ':gem', gem_name])
      config.each do |k, v|
        @mock_database.expect(:hset, nil, [name + ':config', k, v])
      end

      Pepito::Adapters.stub :const_get, mock_adapter_klass do
        Thread.stub :new, nil do
          assert_equal 0, @robot.start_adapter(name, gem_name, config)
          assert mock_adapter.verify
          assert mock_adapter_klass.verify
          assert @mock_database.verify
        end
      end
    end
  end

  def test_stop_adapter
    # Fail to stop if adapter is in the default adapters list
    @robot.default_adapters.stub :include?, true do
      assert_equal 1, @robot.stop_adapter('')
    end

    # Fail to stop if adapter is not running
    @robot.default_adapters.stub :include?, false do
      @robot.adapters.stub :key?, false do
        assert_equal 2, @robot.stop_adapter('')
      end
    end

    # Stop successfully otherwise
    @robot.default_adapters.stub :include?, false do
      @robot.adapters.stub :key?, true do
        @robot.adapters.stub :delete, nil do
          mock_adapter = MiniTest::Mock.new
          mock_adapter.expect(:stop, nil)
          @robot.adapters.stub :[], mock_adapter do
            name = 'adapter'
            @mock_database.expect(:srem, nil, ['adapters', name])
            @mock_database.expect(:del, nil, [name + ':config'])
            @mock_database.expect(:del, nil, [name + ':gem'])

            assert_equal 0, @robot.stop_adapter(name)
            assert mock_adapter.verify
            assert @mock_database.verify
          end
        end
      end
    end
  end

  def test_adapter?
    @robot.adapters.stub :key?, true do
      assert @robot.adapter?('')
    end
    @robot.adapters.stub :key?, false do
      refute @robot.adapter?('')
    end
  end
end
